"""
PyserSSH - A Scriptable SSH server. For more info visit https://github.com/DPSoftware-Foundation/PyserSSH
Copyright (C) 2023-2024 DPSoftware Foundation (MIT)

Visit https://github.com/DPSoftware-Foundation/PyserSSH

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
import shlex

from ..interactive import Send

def are_permissions_met(permission_list, permission_require):
    return set(permission_require).issubset(set(permission_list))

class XHandler:
    def __init__(self, enablehelp=True, showusageonworng=True):
        self.handlers = {}
        self.categories = {}
        self.enablehelp = enablehelp
        self.showusageonworng = showusageonworng
        self.serverself = None
        self.commandnotfound = None

    def command(self, category=None, name=None, aliases=None, permissions: list = None):
        def decorator(func):
            nonlocal name, category
            if name is None:
                name = func.__name__
            command_name = name
            command_description = func.__doc__  # Read the docstring
            parameters = inspect.signature(func).parameters
            command_args = []
            has_args = False
            has_kwargs = False
            
            for param in list(parameters.values())[1:]:  # Exclude first parameter (client)
                if param.kind == inspect.Parameter.VAR_POSITIONAL:
                    has_args = True
                elif param.kind == inspect.Parameter.VAR_KEYWORD:
                    has_kwargs = True
                elif param.default != inspect.Parameter.empty:  # Check if parameter has default value
                    if param.annotation == bool:
                        command_args.append(f"--{param.name}")
                    else:
                        command_args.append((f"{param.name}", param.default))
                else:
                    command_args.append(param.name)
            
            if category is None:
                category = 'No Category'
            if category not in self.categories:
                self.categories[category] = {}
            self.categories[category][command_name] = {
                'description': command_description.strip() if command_description else "",
                'args': command_args,
                "permissions": permissions,
                'has_args': has_args,
                'has_kwargs': has_kwargs
            }
            self.handlers[command_name] = func
            if aliases:
                for alias in aliases:
                    self.handlers[alias] = func
            return func

        return decorator

    def call(self, client, command_string):
        tokens = shlex.split(command_string)
        command_name = tokens[0]
        args = tokens[1:]
        
        if command_name == "help" and self.enablehelp:
            if args:
                Send(client, self.get_help_command_info(args[0]))
            else:
                Send(client, self.get_help_message())
                Send(client, "Type 'help <command>' for more info on a command.")
        else:
            if command_name in self.handlers:
                command_func = self.handlers[command_name]
                command_info = self.get_command_info(command_name)
                if command_info and command_info.get('permissions'):
                    if not are_permissions_met(self.serverself.accounts.get_permissions(client.get_name()), command_info.get('permissions')):
                        Send(client, f"Permission denied. You do not have permission to execute '{command_name}'.")
                        return

                command_args = inspect.signature(command_func).parameters
                final_args = {}
                final_kwargs = {}
                i = 0

                while i < len(args):
                    arg = args[i]
                    if arg.startswith('-'):
                        arg_name = arg.lstrip('-')
                        if arg_name not in command_args:
                            if self.showusageonworng:
                                Send(client, self.get_help_command_info(command_name))
                            Send(client, f"Invalid flag '{arg_name}' for command '{command_name}'.")
                            return
                        if command_args[arg_name].annotation == bool:
                            final_args[arg_name] = True
                            i += 1
                        else:
                            if i + 1 < len(args):
                                final_args[arg_name] = args[i + 1]
                                i += 2
                            else:
                                if self.showusageonworng:
                                    Send(client, self.get_help_command_info(command_name))
                                Send(client, f"Missing value for flag '{arg_name}' for command '{command_name}'.")
                                return
                    else:
                        if command_info['has_args']:
                            final_args.setdefault('args', []).append(arg)
                        elif command_info['has_kwargs']:
                            final_kwargs[arg] = args[i + 1] if i + 1 < len(args) else None
                            i += 1
                        else:
                            if len(final_args) + 1 < len(command_args):
                                param = list(command_args.values())[len(final_args) + 1]
                                final_args[param.name] = arg
                            else:
                                if self.showusageonworng:
                                    Send(client, self.get_help_command_info(command_name))
                                Send(client, f"Unexpected argument '{arg}' for command '{command_name}'.")
                                return
                        i += 1

                # Check for required positional arguments
                for param in list(command_args.values())[1:]:  # Skip client argument
                    if param.name not in final_args and param.default == inspect.Parameter.empty:
                        if self.showusageonworng:
                            Send(client, self.get_help_command_info(command_name))
                        Send(client, f"Missing required argument '{param.name}' for command '{command_name}'")
                        return

                final_args_list = [final_args.get(param.name, param.default) for param in list(command_args.values())[1:]]

                if command_info['has_kwargs']:
                    final_args_list.append(final_kwargs)

                return command_func(client, *final_args_list)
            else:
                if self.commandnotfound:
                    self.commandnotfound(client, command_name)
                    return
                else:
                    Send(client, f"{command_name} not found")
                    return

    def get_command_info(self, command_name):
        found_command = None
        for category, commands in self.categories.items():
            if command_name in commands:
                found_command = commands[command_name]
                break
            else:
                for cmd, cmd_info in commands.items():
                    if 'aliases' in cmd_info and command_name in cmd_info['aliases']:
                        found_command = cmd_info
                        break
                if found_command:
                    break

        if found_command:
            return {
                'name': command_name,
                'description': found_command['description'].strip() if found_command['description'] else "",
                'args': found_command['args'],
                'category': category,
                'permissions': found_command['permissions'],
                'has_args': found_command['has_args'],
                'has_kwargs': found_command['has_kwargs']
            }

    def get_help_command_info(self, command):
        command_info = self.get_command_info(command)
        aliases = command_info.get('aliases', [])
        help_message = f"{command_info['name']}"
        if aliases:
            help_message += f" ({', '.join(aliases)})"
        help_message += "\n"
        help_message += f"{command_info['description']}\n"
        help_message += f"Usage: {command_info['name']}"
        for arg in command_info['args']:
            if isinstance(arg, tuple):
                if isinstance(arg[1], bool):
                    help_message += f" [--{arg[0]}]"
                else:
                    help_message += f" [-{arg[0]} {arg[1]}]"
            else:
                help_message += f" <{arg}>"
        if command_info['has_args']:
            help_message += " [<args>...]"
        if command_info['has_kwargs']:
            help_message += " [--<key>=<value>...]"
        return help_message

    def get_help_message(self):
        help_message = ""
        for category, commands in self.categories.items():
            help_message += f"{category}:\n"
            for command_name, command_info in commands.items():
                help_message += f"  {command_name}"
                if command_info['description']:
                    help_message += f" - {command_info['description']}"
                help_message += "\n"
        return help_message

    def get_all_commands(self):
        all_commands = {}
        for category, commands in self.categories.items():
            all_commands[category] = commands
        return all_commands




"""
PyserSSH - A Scriptable SSH server. For more info visit https://github.com/DPSoftware-Foundation/PyserSSH
Copyright (C) 2023-2024 DPSoftware Foundation (MIT)

Visit https://github.com/DPSoftware-Foundation/PyserSSH

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import socket
import time
import logging

from .sysfunc import replace_enter_with_crlf
from .syscom import systemcommand

logger = logging.getLogger("PyserSSH")

def expect(self, client, echo=True):
    buffer = bytearray()
    cursor_position = 0
    outindexall = 0
    history_index_position = 0  # Initialize history index position outside the loop
    chan = client["channel"]
    peername = client["peername"]

    try:
        while True:
            try:
                byte = chan.recv(1)
            except socket.timeout:
                chan.setblocking(False)
                chan.settimeout(None)
                chan.close()
                raise EOFError()

            self._handle_event("rawtype", self.client_handlers[chan.getpeername()], byte)

            self.client_handlers[chan.getpeername()]["last_activity_time"] = time.time()

            if not byte or byte == b'\x04':
                raise EOFError()
            elif byte == b'\x03':
                pass
            elif byte == b'\t':
                pass
            elif byte == b'\x7f' or byte == b'\x08':
                if cursor_position > 0:
                    buffer = buffer[:cursor_position - 1] + buffer[cursor_position:]
                    cursor_position -= 1
                    outindexall -= 1
                    if cursor_position != outindexall:
                        chan.sendall(b"\b \b")
                        chan.sendall(buffer[cursor_position:])
                    else:
                        chan.sendall(b"\b \b")
                else:
                    chan.sendall(b"\x07")
            elif byte == b"\x1b" and chan.recv(1) == b'[':
                arrow_key = chan.recv(1)
                if not self.disable_scroll_with_arrow:
                    if arrow_key == b'C':
                        # Right arrow key, move cursor right if not at the end
                        if cursor_position < len(buffer):
                            chan.sendall(b'\x1b[C')
                        #    cursor_position += 1
                            cursor_position = min(len(buffer), cursor_position + 1)

                    elif arrow_key == b'D':
                        # Left arrow key, move cursor left if not at the beginning
                        if cursor_position > 0:
                            chan.sendall(b'\x1b[D')
                        #    cursor_position -= 1
                            cursor_position = max(0, cursor_position - 1)

                if self.history:
                    if arrow_key == b'A':
                        if history_index_position == 0:
                            command = self.accounts.get_lastcommand(client["current_user"])
                        else:
                            command = self.accounts.get_history(client["current_user"], history_index_position)

                        # Clear the buffer
                        for i in range(cursor_position):
                            chan.send(b"\b \b")

                        # Update buffer and cursor position with the new command
                        buffer = bytearray(command.encode('utf-8'))
                        cursor_position = len(buffer)
                        outindexall = cursor_position

                        # Print the updated buffer
                        chan.sendall(buffer)

                        history_index_position += 1
                    elif arrow_key == b'B':
                        if history_index_position != -1:
                            if history_index_position == 0:
                                command = self.accounts.get_lastcommand(client["current_user"])
                            else:
                                command = self.accounts.get_history(client["current_user"], history_index_position)

                            # Clear the buffer
                            for i in range(cursor_position):
                                chan.send(b"\b \b")

                            # Update buffer and cursor position with the new command
                            buffer = bytearray(command.encode('utf-8'))
                            cursor_position = len(buffer)
                            outindexall = cursor_position

                            # Print the updated buffer
                            chan.sendall(buffer)
                        else:
                            history_index_position = 0
                            for i in range(cursor_position):
                                chan.send(b"\b \b")

                            buffer.clear()
                            cursor_position = 0
                            outindexall = 0

                        history_index_position -= 1

            elif byte in (b'\r', b'\n'):
                break
            else:
                history_index_position = -1

                self._handle_event("type", self.client_handlers[chan.getpeername()], byte)
                if echo:
                    if outindexall != cursor_position:
                        chan.sendall(b" ")
                        chan.sendall(b'\033[s')
                        chan.sendall(byte + buffer[cursor_position:])
                        chan.sendall(b'\033[u')
                    else:
                        chan.sendall(byte)

                #print(buffer[:cursor_position], byte, buffer[cursor_position:])
                buffer = buffer[:cursor_position] + byte + buffer[cursor_position:]
                cursor_position += 1
                outindexall += 1

            client["inputbuffer"] = buffer

        if echo:
            chan.sendall(b'\r\n')

        command = str(buffer.decode('utf-8')).strip()

        if self.history and command.strip() != "" and self.accounts.get_lastcommand(client["current_user"]) != command:
            self.accounts.add_history(client["current_user"], command)
            client["last_command"] = command

        if command.strip() != "":
            if self.accounts.get_user_timeout(self.client_handlers[chan.getpeername()]["current_user"]) != None:
                chan.setblocking(False)
                chan.settimeout(None)

            try:
                if self.enasyscom:
                    sct = systemcommand(client, command)
                else:
                    sct = False

                if not sct:
                    if self.XHandler != None:
                        self._handle_event("beforexhandler", client, command)

                        self.XHandler.call(client, command)

                        self._handle_event("afterxhandler", client, command)
                    else:
                        self._handle_event("command", client, command)

            except Exception as e:
                self._handle_event("error", client, e)
        if echo:
            try:
                chan.send(replace_enter_with_crlf(client["prompt"] + " "))
            except:
                logger.error("Send error")

        chan.setblocking(False)
        chan.settimeout(None)

        if self.accounts.get_user_timeout(self.client_handlers[chan.getpeername()]["current_user"]) != None:
            chan.setblocking(False)
            chan.settimeout(self.accounts.get_user_timeout(self.client_handlers[chan.getpeername()]["current_user"]))
    except socket.error:
        pass
    except Exception as e:
        logger.error(str(e))
    finally:
        try:
            if not byte:
                logger.info(f"{peername} is disconnected")
                self._handle_event("disconnected", self.client_handlers[peername])
        except:
            logger.info(f"{peername} is disconnected")
            self._handle_event("disconnected", self.client_handlers[peername])


"""
PyserSSH - A Scriptable SSH server. For more info visit https://github.com/DPSoftware-Foundation/PyserSSH
Copyright (C) 2023-2024 DPSoftware Foundation (MIT)

Visit https://github.com/DPSoftware-Foundation/PyserSSH

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import time
import paramiko
import threading
from functools import wraps
import logging
import socket
import random
import traceback

from .system.SFTP import SSHSFTPServer
from .system.sysfunc import replace_enter_with_crlf
from .system.interface import Sinterface
from .system.inputsystem import expect
from .system.info import __version__, system_banner
from .system.clientype import Client as Clientype

# paramiko.sftp_file.SFTPFile.MAX_REQUEST_SIZE = pow(2, 22)

sftpclient = ["WinSCP", "Xplore"]

logger = logging.getLogger("PyserSSH")

class Server:
    def __init__(self, accounts, system_message=True, disable_scroll_with_arrow=True, sftp=False, system_commands=True, compression=True, usexternalauth=False, history=True, inputsystem=True, XHandler=None, title=f"PyserSSH v{__version__}", inspeed=32768, enable_preauth_banner=False, enable_exec_system_command=True, enable_remote_status=False, inputsystem_echo=True):
        """
        system_message set to False to disable welcome message from system
        disable_scroll_with_arrow set to False to enable seek text with arrow (Beta)
        sftp set to True to enable SFTP server
        system_commands set to False to disable system commmands
        compression set to False to disable SSH compression
        enable_remote_status set to True to enable mobaxterm remote monitor (Beta)
        """
        self.sysmess = system_message
        self.accounts = accounts
        self.disable_scroll_with_arrow = disable_scroll_with_arrow
        self.sftpena = sftp
        self.enasyscom = system_commands
        self.compressena = compression
        self.usexternalauth = usexternalauth
        self.history = history
        self.enainputsystem = inputsystem
        self.XHandler = XHandler
        self.title = title
        self.inspeed = inspeed
        self.enaloginbanner = enable_preauth_banner
        self.enasysexec = enable_exec_system_command
        self.enaremostatus = enable_remote_status
        self.inputsysecho = inputsystem_echo

        if self.XHandler != None:
            self.XHandler.serverself = self

        self._event_handlers = {}
        self.client_handlers = {}  # Dictionary to store event handlers for each client
        self.__processmode = None
        self.__serverisrunning = False
        self.__daemon = False

        if self.enasyscom:
            print("\033[33m!!Warning!! System commands is enable! \033[0m")

    def on_user(self, event_name):
        def decorator(func):
            @wraps(func)
            def wrapper(client, *args, **kwargs):
                # Ignore the third argument
                filtered_args = args[:2] + args[3:]
                return func(client, *filtered_args, **kwargs)
            self._event_handlers[event_name] = wrapper
            return wrapper
        return decorator

    def handle_client_disconnection(self, handler, chandlers):
        if not chandlers["channel"].get_transport().is_active():
            if handler:
                handler(chandlers)
            del self.client_handlers[chandlers["peername"]]

    def _handle_event(self, event_name, *args, **kwargs):
        handler = self._event_handlers.get(event_name)
        if event_name == "error" and isinstance(args[0], Clientype):
            args[0].last_error = traceback.format_exc()

        if event_name == "disconnected":
            self.handle_client_disconnection(handler, *args, **kwargs)
        elif handler:
            return handler(*args, **kwargs)

    def handle_client(self, socketchannel, addr):
        self._handle_event("pressh", socketchannel)

        try:
            bh_session = paramiko.Transport(socketchannel)
        except OSError:
            return

        bh_session.add_server_key(self.private_key)

        bh_session.use_compression(self.compressena)

        bh_session.default_window_size = 2147483647
        bh_session.packetizer.REKEY_BYTES = pow(2, 40)
        bh_session.packetizer.REKEY_PACKETS = pow(2, 40)

        bh_session.default_max_packet_size = self.inspeed

        server = Sinterface(self)
        try:
            bh_session.start_server(server=server)
        except:
            return

        logger.info(bh_session.remote_version)

        channel = bh_session.accept()

        if self.sftpena:
            bh_session.set_subsystem_handler('sftp', paramiko.SFTPServer, SSHSFTPServer, channel, self.accounts, self.client_handlers)

        if not bh_session.is_authenticated():
            logger.warning("user not authenticated")
            bh_session.close()
            return

        if channel is None:
            logger.warning("no channel")
            bh_session.close()
            return

        try:
            logger.info("user authenticated")
            peername = bh_session.getpeername()
            if peername not in self.client_handlers:
                # Create a new event handler for this client if it doesn't exist
                self.client_handlers[peername] = Clientype(channel, bh_session, peername)

            client_handler = self.client_handlers[peername]
            client_handler["current_user"] = bh_session.get_username()
            client_handler["channel"] = channel  # Update the channel attribute for the client handler
            client_handler["transport"] = bh_session  # Update the channel attribute for the client handler
            client_handler["last_activity_time"] = time.time()
            client_handler["last_login_time"] = time.time()
            client_handler["prompt"] = self.accounts.get_prompt(bh_session.get_username())
            client_handler["session_id"] = random.randint(10000, 99999) + int(time.time() * 1000)

            self.accounts.set_user_last_login(self.client_handlers[channel.getpeername()]["current_user"], peername[0])

            logger.info("saved user data to client handlers")

            #if not any(bh_session.remote_version.split("-")[2].startswith(prefix) for prefix in sftpclient):
            if int(channel.out_window_size) != int(bh_session.default_window_size):
                logger.info("user is ssh")
                #timeout for waiting 10 sec
                for i in range(100):
                    if self.client_handlers[channel.getpeername()]["windowsize"]:
                        break
                    time.sleep(0.1)

                if self.client_handlers[channel.getpeername()]["windowsize"] == {}:
                    logger.info("timeout for waiting window size in 10 sec")
                    self.client_handlers[channel.getpeername()]["windowsize"] = {
                        "width": 80,
                        "height": 24,
                        "pixelwidth": 0,
                        "pixelheight": 0
                    }

                try:
                    self._handle_event("pre-shell", self.client_handlers[channel.getpeername()])
                except Exception as e:
                    self._handle_event("error", self.client_handlers[channel.getpeername()], e)

                while self.client_handlers[channel.getpeername()]["isexeccommandrunning"]:
                    time.sleep(0.1)

                userbanner = self.accounts.get_banner(self.client_handlers[channel.getpeername()]["current_user"])

                if self.accounts.get_user_enable_inputsystem_echo(self.client_handlers[channel.getpeername()]["current_user"]) and self.inputsysecho:
                    echo = True
                else:
                    echo = False

                if echo:
                    if self.title.strip() != "":
                        channel.send(f"\033]0;{self.title}\007".encode())

                    if self.sysmess or userbanner != None:
                        if userbanner is None and self.sysmess:
                            channel.sendall(replace_enter_with_crlf(system_banner))
                        elif userbanner != None and self.sysmess:
                            channel.sendall(replace_enter_with_crlf(system_banner))
                            channel.sendall(replace_enter_with_crlf(userbanner))
                        elif userbanner != None and not self.sysmess:
                            channel.sendall(replace_enter_with_crlf(userbanner))

                        channel.sendall(replace_enter_with_crlf("\n"))

                client_handler["connecttype"] = "ssh"

                try:
                    self._handle_event("connect", self.client_handlers[channel.getpeername()])
                except Exception as e:
                    self._handle_event("error", self.client_handlers[channel.getpeername()], e)

                if self.enainputsystem and self.accounts.get_user_enable_inputsystem(self.client_handlers[channel.getpeername()]["current_user"]):
                    try:
                        if self.accounts.get_user_timeout(self.client_handlers[channel.getpeername()]["current_user"]) != None:
                            channel.setblocking(False)
                            channel.settimeout(self.accounts.get_user_timeout(self.client_handlers[channel.getpeername()]["current_user"]))

                        if echo:
                            channel.send(replace_enter_with_crlf(self.client_handlers[channel.getpeername()]["prompt"] + " "))

                        while True:
                            expect(self, self.client_handlers[channel.getpeername()], echo)
                    except KeyboardInterrupt:
                        self._handle_event("disconnected", self.client_handlers[peername])
                        channel.close()
                        bh_session.close()
                    except Exception as e:
                        self._handle_event("error", client_handler, e)
                        logger.error(e)
                    finally:
                        self._handle_event("disconnected", self.client_handlers[peername])
                        channel.close()
            else:
                if self.sftpena:
                    logger.info("user is sftp")
                    if self.accounts.get_user_sftp_allow(self.client_handlers[channel.getpeername()]["current_user"]):
                        client_handler["connecttype"] = "sftp"
                        self._handle_event("connectsftp", self.client_handlers[channel.getpeername()])
                        while bh_session.is_active():
                            time.sleep(0.1)

                        self._handle_event("disconnected", self.client_handlers[peername])

                    else:
                        self._handle_event("disconnected", self.client_handlers[peername])
                        channel.close()
                else:
                    self._handle_event("disconnected", self.client_handlers[peername])
                    channel.close()
        except:
            bh_session.close()

    def stop_server(self):
        logger.info("Stopping the server...")
        try:
            for client_handler in self.client_handlers.values():
                channel = client_handler.channel
                if channel:
                    channel.close()
            self.__serverisrunning = False
            self.server.close()

            logger.info("Server stopped.")
        except Exception as e:
            logger.error(f"Error occurred while stopping the server: {e}")

    def _start_listening_thread(self):
        try:
            logger.info("Start Listening for connections...")
            while self.__serverisrunning:
                client, addr = self.server.accept()
                if self.__processmode == "thread":
                    client_thread = threading.Thread(target=self.handle_client, args=(client, addr), daemon=True)
                    client_thread.start()
                else:
                    self.handle_client(client, addr)
                time.sleep(1)
        except KeyboardInterrupt:
            self.stop_server()
        except Exception as e:
            logger.error(e)

    def run(self, private_key_path=None, host="0.0.0.0", port=2222, mode="thread", maxuser=0, daemon=False):
        """mode: single, thread
        protocol: ssh, telnet
        """
        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        self.server.bind((host, port))

        if private_key_path != None:
            self.private_key = paramiko.RSAKey(filename=private_key_path)
        else:
            raise ValueError("No private key")

        if maxuser == 0:
            self.server.listen()
        else:
            self.server.listen(maxuser)

        self.__processmode = mode.lower()
        self.__serverisrunning = True
        self.__daemon = daemon

        client_thread = threading.Thread(target=self._start_listening_thread, daemon=self.__daemon)
        client_thread.start()




"""
PyserSSH - A Scriptable SSH server. For more info visit https://github.com/DPSoftware-Foundation/PyserSSH
Copyright (C) 2023-2024 DPSoftware Foundation (MIT)

Visit https://github.com/DPSoftware-Foundation/PyserSSH

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re

from ..interactive import Clear, Send, wait_inputkey
from ..system.sysfunc import text_centered_screen

class TextDialog:
    def __init__(self, client, content="", title=""):
        self.client = client

        self.windowsize = client["windowsize"]
        self.title = title
        self.content = content

    def render(self):
        Clear(self.client)
        Send(self.client, self.title)
        Send(self.client, "-" * self.windowsize["width"])

        generatedwindow = text_centered_screen(self.content, self.windowsize["width"], self.windowsize["height"]-3, " ")

        Send(self.client, generatedwindow)

        Send(self.client, "Press 'enter' to continue", ln=False)

        self.waituserenter()

    def waituserenter(self):
        while True:
            if wait_inputkey(self.client, raw=True) == b'\r':
                Clear(self.client)
                break
            pass

class MenuDialog:
    def __init__(self, client, choose: list, title="", desc=""):
        self.client = client

        self.title = title
        self.choose = choose
        self.desc = desc
        self.contentallindex = len(choose) - 1
        self.selectedindex = 0
        self.selectstatus = 0 # 0 none 1 selected 2 cancel

    def render(self):
        tempcontentlist = self.choose.copy()

        Clear(self.client)
        Send(self.client, self.title)
        Send(self.client, "-" * self.client["windowsize"]["width"])

        tempcontentlist[self.selectedindex] = "> " + tempcontentlist[self.selectedindex]

        exported = "\n".join(tempcontentlist)

        if not self.desc.strip() == "":
            contenttoshow = (
                f"{self.desc}\n\n"
                f"{exported}"
            )
        else:
            contenttoshow = (
                f"{exported}"
            )

        generatedwindow = text_centered_screen(contenttoshow, self.client["windowsize"]["width"], self.client["windowsize"]["height"]-3, " ")

        Send(self.client, generatedwindow)

        Send(self.client, "Use arrow up/down key to choose and press 'enter' to select or 'c' to cancel", ln=False)

        self._waituserinput()

    def _waituserinput(self):
        keyinput = wait_inputkey(self.client, raw=True)

        if keyinput == b'\r':  # Enter key
            Clear(self.client)
            self.selectstatus = 1
        elif keyinput == b'c':  # 'c' key for cancel
            Clear(self.client)
            self.selectstatus = 2
        elif keyinput == b'\x1b[A':  # Up arrow key
            self.selectedindex -= 1
            if self.selectedindex < 0:
                self.selectedindex = 0
        elif keyinput == b'\x1b[B':  # Down arrow key
            self.selectedindex += 1
            if self.selectedindex > self.contentallindex:
                self.selectedindex = self.contentallindex

        if self.selectstatus == 2:
            self.output()
        elif self.selectstatus == 1:
            self.output()
        else:
            self.render()

    def output(self):
        if self.selectstatus == 2:
            return None
        elif self.selectstatus == 1:
            return self.selectedindex

class TextInputDialog:
    def __init__(self, client, title="", inputtitle="Input Here", password=False):
        self.client = client

        self.title = title
        self.inputtitle = inputtitle
        self.ispassword = password

        self.inputstatus = 0 # 0 none 1 selected 2 cancel
        self.buffer = bytearray()
        self.cursor_position = 0

    def render(self):
        Clear(self.client)
        Send(self.client, self.title)
        Send(self.client, "-" * self.client["windowsize"]["width"])

        if self.ispassword:
            texts = (
                f"{self.inputtitle}\n\n"
                "> " + ("*" * len(self.buffer.decode('utf-8')))
            )
        else:
            texts = (
                f"{self.inputtitle}\n\n"
                "> " + self.buffer.decode('utf-8')
            )

        generatedwindow = text_centered_screen(texts, self.client["windowsize"]["width"], self.client["windowsize"]["height"]-3, " ")

        Send(self.client, generatedwindow)

        Send(self.client, "Press 'enter' to select or 'ctrl+c' to cancel", ln=False)

        self._waituserinput()

    def _waituserinput(self):
        keyinput = wait_inputkey(self.client, raw=True)

        if keyinput == b'\r':  # Enter key
            Clear(self.client)
            self.inputstatus = 1
        elif keyinput == b'\x03':  # 'ctrl + c' key for cancel
            Clear(self.client)
            self.inputstatus = 2

        try:
            if keyinput == b'\x7f' or keyinput == b'\x08':  # Backspace
                if self.cursor_position > 0:
                    # Move cursor back, erase character, move cursor back again
                    self.buffer = self.buffer[:self.cursor_position - 1] + self.buffer[self.cursor_position:]
                    self.cursor_position -= 1
            elif bool(re.compile(b'\x1b\[[0-9;]*[mGK]').search(keyinput)):
                pass
            else:  # Regular character
                self.buffer = self.buffer[:self.cursor_position] + keyinput + self.buffer[self.cursor_position:]
                self.cursor_position += 1
        except Exception:
            raise

        if self.inputstatus == 2:
            self.output()
        elif self.inputstatus == 1:
            self.output()
        else:
            self.render()

    def output(self):
        if self.inputstatus == 2:
            return None
        elif self.inputstatus == 1:
            return self.buffer.decode('utf-8')



"""
PyserSSH - A Scriptable SSH server. For more info visit https://github.com/DPSoftware-Foundation/PyserSSH
Copyright (C) 2023-2024 DPSoftware Foundation (MIT)

Visit https://github.com/DPSoftware-Foundation/PyserSSH

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import time
from paramiko.transport import Transport
from paramiko.channel import Channel

class Client:
    def __init__(self, channel, transport, peername):
        self.current_user = None
        self.transport: Transport = transport
        self.channel: Channel = channel
        self.subchannel = {}
        self.connecttype = None
        self.last_activity_time = None
        self.last_login_time = None
        self.windowsize = {}
        self.x11 = {}
        self.prompt = None
        self.inputbuffer = None
        self.peername = peername
        self.auth_method = self.transport.auth_handler.auth_method
        self.session_id = None
        self.terminal_type = None
        self.env_variables = {}
        self.last_error = None
        self.last_command = None
        self.isexeccommandrunning = False

    def get_id(self):
        return self.session_id

    def get_name(self):
        return self.current_user

    def get_peername(self):
        return self.current_user

    def get_prompt(self):
        return self.prompt

    def get_channel(self):
        return self.channel

    def get_prompt_buffer(self):
        return str(self.inputbuffer)

    def get_terminal_size(self):
        return self.windowsize["width"], self.windowsize["height"]

    def get_connection_type(self):
        return self.connecttype

    def get_auth_with(self):
        return self.auth_method

    def get_session_duration(self):
        return time.time() - self.last_login_time

    def get_environment(self, variable):
        return self.env_variables[variable]

    def get_last_error(self):
        return self.last_error

    def get_last_command(self):
        return self.last_command

    def set_name(self, name):
        self.current_user = name

    def set_prompt(self, prompt):
        self.prompt = prompt

    def set_environment(self, variable, value):
        self.env_variables[variable] = value

    def open_new_subchannel(self, timeout=None):
        try:
            channel = self.transport.accept(timeout)
            id = channel.get_id()
        except:
            return None, None

        self.subchannel[id] = channel
        return id, channel

    def get_subchannel(self, id):
        return self.subchannel[id]

    def switch_user(self, user):
        self.current_user = user
        self.transport.auth_handler.username = user

    def close_subchannel(self, id):
        self.subchannel[id].close()

    def close(self):
        self.channel.close()

    # for backward compatibility only
    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __str__(self):
        return f"client id: {self.session_id}"

    def __repr__(self):
        # Get the dictionary of instance attributes
        attrs = vars(self)  # or self.__dict__

        # Filter out attributes that are None
        non_none_attrs = {key: value for key, value in attrs.items() if value is not None}

        # Build a string representation
        attrs_repr = ', '.join(f"{key}={value!r}" for key, value in non_none_attrs.items())
        return f"Client({attrs_repr})"